/*
**    Mesh Converter
**    Copyright (C) 2008 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



/*
** Mesh_object.h
** 
** Made by damien
** Login   <damien@ortie>
** 
** Started on  Fri Jun  6 19:26:06 2008 damien
** Last update Fri Jun  6 19:26:06 2008 damien
*/

#ifndef   	MESH_OBJECT_H_
# define   	MESH_OBJECT_H_


#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <cmath>
#include <string>
#include <errno.h>
#include <stdio.h>

#include <int_vector.h>
#include <tinyXml/tinyxml.h>
#include <tinyXml/tinystr.h>

#include <Connectivity.h>
#include <Point_set.h>
#include <Vector_parameter.h>
#include <Polygon.h>
#include <Curve_3D.h>



/// The mesh object class
/** The C0 mesh associating geometry as a point set and a connectivity */

class Mesh_object
{

public:

  //************************************************//
  //************************************************//
  //CONSTRUCTORS
  //************************************************//
  //************************************************//

  /// Classical constructor
  Mesh_object();
  /// Copy Constructor
  Mesh_object(const Mesh_object& );
  /// Classical destroy
  ~Mesh_object();


  /// set all vector to 0
  int destroy();






  //************************************************//
  //************************************************//
  //Geometry
  //************************************************//
  //************************************************//


  //Get
  //************************************************//

  /// Number of vertices
  /** Number of stored vertex ((x0,y0,z0),...,(xN,yN,zN)) */
  int get_vertex_number() const;

  /// Get vertex value
  double& get_vertex(int k_vertex,int k_dim);
  /// Get vertex value
  double& get_vertex(int k_polygon,int k_vertex,int k_dim);
  /// Get vertex value
  double  get_vertex(int k_vertex,int k_dim) const;
  /// Get vertex value
  double  get_vertex(int k_polygon,int k_vertex,int k_dim) const;
  /// Get vertex pointer in double*
  const double* get_vertex_p(int k_vertex);
  /// Get vertex pointer in double*
  const double* get_vertex_p(int k_polygon,int k_vertex);
  /// Get vertex and return V_3D class
  V_3D get_vertex_v3d(int k_vertex) const;
  /// Get vertex and return V_3D class
  V_3D get_vertex_v3d(int k_polygon,int k_vertex) const;


  //Set
  //************************************************//

  /// Set vertex from the 3 coordinates
  int set_vertex(int k_vertex,double x,double y,double z);
  /// Set vertex from pointing a coordinate
  int set_vertex(int k_vertex,int k_dim,double value);
  /// Set vertex form V_3D class
  int set_vertex(int k_vertex,V_3D v);
  /// Set vertex from pointer (only one point)=> assume size(v)=3
  int set_vertex(int k_vertex,double* v);
  /// Set vertex from a set of point in a pointer mode size(v)=3*dim
  int set_vertex(int k_vertex,int dim,double *v);
  /// Set vertex from Point_Set
  int set_vertex(const Point_set& p);

  //ADD 
  //************************************************//

  /// Add a given vertex
  int add_vertex(double x,double y,double z);
  /// Add a given vertex
  int add_vertex(const V_3D& v);
  /// Add a vector of vertex
  int add_vertex(const std::vector <double>& to_add);
  /// Add a vector of vertex in a V_3D vector
  int add_vertex(const std::vector <V_3D>& to_add);
  /// Add a vertex via pointer (size should be 3)
  int add_vertex(const double* v);
  /// Add multiple vertices (size(v) should be 3*dim)
  int add_vertex(int dim,double *v);
  /// Add from a Point_set class
  int add_vertex(const Point_set& p);



  //ADD/DELETE UNIQUE
  //************************************************//

  /// Add unique vertex with an error epsilon (return -1 if not found)
  int add_unique(double x,double y,double z,double epsilon);
  /// Add unique vertex with an error epsilon (return -1 if not found)
  int add_unique(const V_3D& v,double epsilon);
  /// Add unique vertex with an error epsilon (return -1 if not found)
  int add_unique(const double* v,double epsilon);

  /// look if a given vertex exists (return -1 if not)
  int exist(const V_3D& v) const;
  /// look if a given vertex exists (return -1 if not)
  int exist(const V_3D& v,double epsilon) const;
  /// look if a given vertex exists (return -1 if not)
  int exist(double x,double y,double z) const;
  /// look if a given vertex exists (return -1 if not)
  int exist(double x,double y,double z,double epsilon) const;


  /// Eliminate a given vertex
  int eliminate(int k_index);
  /// Eliminate a set of vertices
  int eliminate(const std::vector <int>& k_index);
  /// Eliminate a set of vertices
  int eliminate(const std::vector <double>& k_index);
  /// Eliminate a vertex equal at (x,y,z)
  int eliminate(double x,double y,double z);
  /// Eliminate a vertex equal at (x,y,z) with the given error=epsilon
  int eliminate(double x,double y,double z,double epsilon);
  /// Eliminate a vertex such that ||vertex-x||<=epsilon
  int eliminate(const V_3D& x,double epsilon);
  /// Eliminate a vertex such that ||vertex-x|| is small enough
  int eliminate(const V_3D& x);
  /// Eliminate a vertex such that ||vertex-x||<=epsilon
  int eliminate(const double* x,double epsilon);
  /// Eliminate a vertex such that ||vertex-x|| is small enough
  int eliminate(const double* x);





  //TRANSFORMATIONS
  //************************************************//

  /// Apply a general 4x4 matrix of transformation to every vertex
  int matrix_transform(const Matrix& M);

  /// Rotate every vertex by a rotation of angle around the vector n
  int rotate(const V_3D& n,double angle);
  /// Rotate every vertex by a rotation of angle around the vector (nx,ny,nz)
  int rotate(double n_x,double n_y,double n_z,double angle);
  /// Rotate every vertex by a rotation of angle around the vector n
  int rotate(const double *n,double angle);

  /// Rotate every vertex by a rotation of angle around the vector n centered to the barycenter of the points set
  int rotate_centered(const V_3D& n,double angle);
  /// Rotate every vertex by a rotation of angle around the vector (nx,ny,nz) centered to the barycenter of the points set
  int rotate_centered(double n_x,double n_y,double n_z,double angle);
  /// Rotate every vertex by a rotation of angle around the vector n centered to the barycenter of the points set
  int rotate_centered(const double *n,double angle);

  /// Rotate every vertex by a rotation of angle around the vector n centered on the point: center
  int rotate(const V_3D& n,double angle,const V_3D& center);
  /// Rotate every vertex by a rotation of angle around the vector (nx,ny,nz) centered on the point: center
  int rotate(double n_x,double n_y,double n_z,double angle,const V_3D& center);
  /// Rotate every vertex by a rotation of angle around the vector n centered on the point: center
  int rotate(const double *n,double angle,const V_3D& center);



  /// Translate every vertex by the vector (tx,ty,tz)
  int translate(double t_x,double t_y,double t_z);
  /// Translate every vertex by the vector t
  int translate(const double *t);
  /// Translate every vertex by the vector t
  int translate(const V_3D& t);


  /// Scale every vertex by the factor (sx,sy,sz)
  int scale(double s_x,double s_y,double s_z);
  /// Scale every vertex by the factor (s[0],s[1],s[2])
  int scale(const double *s);
  /// Scale every vertex by the factor (s[0],s[1],s[2])
  int scale(const V_3D& s);
  /// Scale every vertex by the factor s (the same in every direction)
  int scale(double s);


  /// Scale every vertex by the factor (sx,sy,sz) centered on the barycenter of the mesh
  int scale_centered(double s_x,double s_y,double s_z);
  /// Scale every vertex by the factor (s[0],s[1],s[2]) centered on the barycenter of the mesh
  int scale_centered(const double *s);
  /// Scale every vertex by the factor (s[0],s[1],s[2]) centered on the barycenter of the mesh
  int scale_centered(const V_3D& s);
  /// Scale every vertex by the factor s (the same in every direction) centered on the barycenter of the mesh
  int scale_centered(double s);



  //SIZE
  //************************************************//


  /// Return the middle of the Points Set
  V_3D get_middle_point() const;
  /// Return the min of the Points Set
  V_3D get_min_point() const;
  /// Return the max of the Points Set
  V_3D get_max_point() const;
  /// Return the size of the mesh
  V_3D get_size_mesh() const;

  //center the point set
  int center_mesh();
  //set max size to 1
  int unitize_size();

  

  //FURTHER
  //************************************************//

  /// get the index which is the closest of (x,y,z)
  /** return the distance to (x,y,z) and the index of the closest vertex in *index */
  double get_closest_distance(double x,double y,double z,int *index) const;
  /// get the index which is the closest of x
  /** return the distance to x and the index of the closest vertex in *index */
  double get_closest_distance(const double* x,int *index) const;
  /// get the index which is the closest of x
  /** return the distance to x and the index of the closest vertex in *index */
  double get_closest_distance(const V_3D& x,int *index) const;


  /// get the index which is the further of (x,y,z)
  /** return the distance to (x,y,z) and the index of the further vertex in *index */
  double get_further_distance(double x,double y,double z,int *index) const;
  /// get the index which is the further of x
  /** return the distance to x and the index of the further vertex in *index */
  double get_further_distance(const double* x,int *index) const;
  /// get the index which is the further of x
  /** return the distance to x and the index of the further vertex in *index */
  double get_further_distance(const V_3D& x,int *index) const;

  /// get the distance of the point (x,y,z) to the vertex at position=index
  double distance_to_index(double x,double y,double z,int index) const;
  /// get the distance of the point x to the vertex at position=index
  double distance_to_index(const double* x,int index) const;
  /// get the distance of the point x to the vertex at position=index
  double distance_to_index(const V_3D& x,int index) const;


  /// get the N-closest vertices of the point center
  std::vector <int> N_closest(const V_3D& center,int N_closest) const;
  /// get the N-closest vertices of the point (x,y,z)
  std::vector <int> N_closest(double x,double y,double z,int N_closest) const;
  /// get the N-closest vertices of the point x
  std::vector <int> N_closest(double *x,int N_closest) const;

  /// get the closest vertex
  V_3D get_closest_vertex(const V_3D& center) const;
  /// get the closest vertex
  V_3D get_closest_vertex(double x,double y,double z) const;
  /// get the closest vertex
  V_3D get_closest_vertex(const double *x) const;



  




  //************************************************//
  //************************************************//
  //Connectivity
  //************************************************//
  //************************************************//


  //Get
  //************************************************//

  /// get index from k_index
  int get_index(int k_index) const;
  /// get index from k_index
  int& get_index(int k_index);
  /// get index from k_index of k_polygon
  int get_index(int k_polygon,int k_index) const;
  /// get index from k_index of k_polygon
  int& get_index(int k_polygon,int k_index);
  /// get full_index of the given polygon
  std::vector <int> get_index_of_polygon(int k_polygon) const;
  /// get_pointer
  const int* get_index() const;

  /// get the Connectivity class
  Connectivity& get_connectivity();
  /// get the Connectivity class
  const Connectivity& get_connectivity() const;

  /// size of the vector
  int size_index() const;
  /// number of polygons
  int polygon_number() const;
  /// get size of the given polygon
  int polygon_size(int k_polygon) const;
  /// get number of vertex by taking the max of every vertex index
  int vertex_number() const;


  /// get index of the triangulated version
  int get_index_triangulated(int k_index) const;
  /// get index of the triangulated version
  int& get_index_triangulated(int k_index);
  /// get index of this triangle in the triangulated version
  int get_index_triangulated(int k_triangle,int k_index) const;
  /// get index of this triangle in the triangulated version
  int& get_index_triangulated(int k_triangle,int k_index);
  /// get pointer of the triangulation
  const int* get_index_triangulated() const;
  /// get number of triangules in the triangulated version
  int get_triangle_number() const;

  /// try to find a given polygon return polygon number or -1 if not found
  int find_polygon(const std::vector <int>& polygon) const;

  /// rebuild the index
  /** Call build_all_index in Connectivity */
  int build_connectivity_index();

  /// Triangulate the mesh
  int triangulate();

  /// Get the polygon access from the connectivity
  /**
     Ordered such that for instance
     index0,index1,index2,-1,index3,index4,index5,index6,-1,index7,...
     => access(0)=0, access(1)=4, access(2)=9, ...
   */
  int get_polygon_access(int k_polygon) const;

  
  //Add/Delete polygon
  //************************************************//

  /// add a full polygon
  int add_polygon(const std::vector <int>& polygon);
  /// add a full quad
  int add_polygon(int index_0,int index_1,int index_2,int index_3);
  /// add a full triangle
  int add_polygon(int index_0,int index_1,int index_2);
  /// add index of polygon
  /** does not ensure that the polygon end by -1 (should be manually included)*/
  int add_index(int index);
  /// set polygon number k
  /** slow, needs the re-calculation of the index 
      Also assume that the access is already set-up */
  int set_polygon(int k_polygon,const std::vector <int>& polygon);
  /// set index
  /** should not be use, direct access to the index, does not ensure
      its correct with size */
  int set_index(int k_polygon,int k_index,int index);
  /// delete polygon
  /** assume that all the index are already calculated.
      Slow, will recalculate everything at each time */
  int delete_polygon(int k_polygon);
  /// delete polygon from a given polygon
  int delete_given_polygon(const std::vector <int> polygon);

  /// Flip a given polygon
  /** Inverse the sens of the polygon */
  int flip_polygon(int k_polygon);
  /// Flip every polygon
  int flip_polygon();



  //One Ring
  //************************************************//

  /// get one ring
  int_vector get_one_ring(int k_polygon,int k_vertex) const;
  /// get one ring
  int_vector get_one_ring(int k_vertex) const;
  /// get Polygon ring
  int_vector get_polygon_ring(int k_polygon,int k_vertex) const;
  /// get Polygon ring
  int_vector get_polygon_ring(int k_vertex) const;




  //Edges and Manifold
  //************************************************//

  /// build manifold if there is triangle inversion
  /** Only works if there is repeted index between two triangles (not
      in the correct turning sense */
  int build_manifold();

  /// build entire edges from polygons
  /** Works on 2-manifold */
  int build_edges();


  //************************************************//
  //************************************************//
  //Normal
  //************************************************//
  //************************************************//

  /// build a normal for each polygon
  int build_normal_per_polygon();
  /// build a different normal for every vertex of every polygon
  /** One vertex and one polygon can have multiple normals */
  int build_normal_intermediate(double angle_threshold);
  /// build a smooth normal for every vertex
  int build_normal_per_vertex();
  /// return the type of calculated normal (-1 if not)
  int normal_type() const;
  /// return the normal if they is one normal per polygon (type=1)
  V_3D get_polygon_normal(int k_polygon) const;
  /// return the normal if they is one normal per polygon per vertex (type=3)
  V_3D get_polygon_and_vertex_normal(int k_polygon,int k_vertex) const;
  /// return the normal if they is one normal per vertex (type=1)
  V_3D get_vertex_normal(int k_vertex) const;
  /// set the normal per vertex (if they is one normal per vertex: type=1)
  int set_vertex_normal(int k_vertex,const V_3D& n);
  /// return the normal if they is one normal per vertex (type=1)
  V_3D get_vertex_normal(int k_polygon,int k_vertex) const;
  /// set the normal per vertex (if they is one normal per vertex: type=1)
  int set_vertex_normal(int k_polygon,int k_vertex,const V_3D& n);


  //************************************************//
  //************************************************//
  //Color
  //************************************************//
  //************************************************//

  /// Set up the vector parameter of color for a per vertex color
  int set_per_vertex_color();
  /// Set up the vector parameter of color for a per polygon color
  int set_per_polygon_color();
  /// Set up the vector parameter of color for a per polygon_and_vertex color
  int set_per_polygon_and_vertex_color();


  /// Get the type of color
  /** return -1 if no color
              1 per vertex
	      2 per polygon
	      3 per vertex and polygon
  */
  int color_type() const;

  /// Fill the whole object with a given color
  /** Do by default a per vertex color*/
  int fill_color(double r,double g,double b,double t);

  /// Set per polygon color
  int set_per_polygon_color(int k_polygon,double r,double g,double b,double t);
  /// Get per polygon color
  int get_per_polygon_color(int k_polygon,double *r,double *g,double *b,double *t) const;

  /// Set per vertex color (direct access)
  int set_per_vertex_color(int k_vertex,double r,double g,double b,double t);
  /// Get per vertex color (direct access)
  int get_per_vertex_color(int k_vertex,double *r,double *g,double *b,double *t) const;
  /// Set per vertex color
  int set_per_vertex_color(int k_polygon,int k_vertex,double r,double g,double b,double t);
  /// Get per vertex color
  int get_per_vertex_color(int k_polygon,int k_vertex,double *r,double *g,double *b,double *t) const;



  /// Set per polygon_and_vertex color
  int set_per_polygon_and_vertex_color(int k_polygon,int k_vertex,double r,double g,double b,double t);
  /// Get per polygon_and_vertex color
  int get_per_polygon_and_vertex_color(int k_polygon,int k_vertex,double *r,double *g,double *b,double *t) const;


  /// increase the transparency
  int increase_alpha(double d_alpha);
  /// decrease the transparency
  int decrease_alpha(double d_alpha);


  //************************************************//
  //************************************************//
  //Texture
  //************************************************//
  //************************************************//


  /// enable the intermediate texture (vector of texture and index)
  int set_intermediate_texture();
  /// return if texture is enabled or not
  /** -1: no texture
       0: intermediate texture (vector+index)
   */
  int texture_type() const;
  /// get texture coords
  int get_intermediate_texture(int k_polygon,int k_vertex,double* t0,double* t1) const;
  /// set texture coords
  int set_intermediate_texture(int k_polygon,int k_vertex,double t0,double t1);
  /// Add 2-coords texture in the vector
  int add_intermediate_texture(double t0,double t1);
  /// Add polygon index in the texture vector
  int add_polygon_texture(const std::vector <int>& index);

  /// set the texture number
  int set_texture_number(int k_texture);
  /// get the texture number
  int get_texture_number() const;

  /// get the number of polygon of texture loaded (should be equal to the polygon of the mesh)
  int get_texture_polygon_number() const;

  /// check if texture are [ok] to be displayed
  int check_texture() const;


  //************************************************//
  //************************************************//
  //Polygon Geometry
  //************************************************//
  //************************************************//

  /// Add the polygon in a unique sens
  /** Add the connectivity and the vertex if it does not exists yet. 
      Return a vector containing the n. of connectivity.
   */
  std::vector <int> add_unique_polygon(const Polygon &polygon);


  /// return the designated polygon
  Polygon get_polygon(int k_polygon) const;
  
  /// find the closest point on the mesh surface
  /** find the real closest point by checking every polygons*/
  V_3D closest_mesh_point(const V_3D& x) const;

  /// Get the intersection between the mesh and a plane
  /** The plane is defined by its normal n, and a point x0, so its equation is <n,x-x0>=0
      Return the curve that intersect the mesh. This curve might be
      empty if no intersection. 
      It might also be many curves in cases of non convex mesh, or non
      connected parts
  */
  std::vector <Curve_3D> plane_intersection(const V_3D& n,const V_3D& x0) const;
  

  /// Return the intersection of the Mesh and the half space defined by the oriented plane <n,x-x0>=0
  /** 3 Possibilies for type: 0 - unchanged Mesh
                              1 - partly cutted polygon
			      2 - no more Mesh at all
  */
  Mesh_object half_space_intersection(const V_3D& n,const V_3D& x0,int *type);

  /// Return the intersection of the Mesh and the half space defined by the oriented plane <n,x-x0>=0
  /** Same as half_space_intersection with type*/
  Mesh_object half_space_intersection(const V_3D& n,const V_3D& x0);



  /// Smooth a mesh using a Laplacien operator on the one ring
  /* x = x-lambda*x where v is the vertex going to the one ring barycenter */
  int Laplacien_smooth(double lambda);

  //************************************************//
  //************************************************//
  //Specific Meshes
  //************************************************//
  //************************************************//

  /// build a unit cube from [0,1]x[0,1]x[0,1]
  int build_cube();





  //************************************************//
  //************************************************//
  //Subdivision process
  //************************************************//
  //************************************************//

  /// subdivide the mesh by linking consecutives middle points of the edges
  int subdivide_mid_edge();
  /// subdivide the mesh by linking consecutives middle points of the edges to the barycenter
  int subdivide_barycenter_mid_edge();
  /// subdivide the mesh with the middle edge. If it is a triangle, then 
  int subdivide_mixed_mid_edge();

  /// subdivide the mesh by linking consecutives middle points of the edges
  /** fast but uses more memory in recording every edges twice */
  int fast_subdivide_mid_edge();
  /// subdivide the mesh by linking consecutives middle points of the edges to the barycenter
  /** fast but uses more memory in recording every edges twice */
  int fast_subdivide_barycenter_mid_edge();
  /// subdivide the mesh by mixing the mid edge and the barycenter method for quads and triangles.
  /** fast but uses more memory in recording every edges twice 
      Do barycenter method for quads
      Do mid edges for other polygons (triangles)
   */
  int fast_subdivide_mixed_mid_edge();


  //************************************************//
  //************************************************//
  //INPUT/OUTPUT
  //************************************************//
  //************************************************//

  /// Load a OFF file
  int load_off_file(const char* filename);
  /// Load a OFF file
  int load_off_file(const std::string& filename);
  /// Save Mesh as a OFF file
  int save_off_file(const char* filename) const;
  /// Save Mesh as a OFF file
  int save_off_file(const std::string& filename) const;

  /// Load a OBJ file
  int load_obj_file(const char* filename);
  /// Load a OBJ file
  int load_obj_file(const std::string& filename);
  /// Save Mesh as a OBJ file
  int save_obj_file(const char* filename);
  /// Save Mesh as a OBJ file
  int save_obj_file(const std::string& filename);

  /// Load a COLLADA file
  /** Load every polygons of the file */
  int load_collada_file(const char* filename);
  /// Load a COLLADA file
  /** Load every polygons of the file */
  int load_collada_file(const std::string& filename);

  /// Load a COLLADA file
  /** Load every polygons of the file */
  int load_collada_file(const char* filename,const char* geometry_name);
  /// Load a COLLADA file
  /** Load every polygons of the file */
  int load_collada_file(const std::string& filename,const std::string& geometry_name);


  /// Load a g file
  int load_g_file(const char* filename);
  /// Load a g file
  int load_g_file(const std::string& filename);
  /// save a g file
  int save_g_file(const char* filename) const;
  /// save a g file
  int save_g_file(const std::string& filename) const;

  //************************************************//
  //************************************************//
  //NAME
  //************************************************//
  //************************************************//

  /// set the mesh name
  int set_name(const char* name);
  /// set the mesh name
  int set_name(const std::string& name);
  /// get the mesh name
  std::string get_name() const;
  /// get the mesh name
  std::string& get_name();
  


private:

  /// The connectivity of the mesh
  Connectivity connectivity;
  /// The geometry of the mesh
  Point_set point_set;

  /// Vector of parameters
  /** 1st: normal
      2st: color
      3rd: texture
   */
  std::vector <Vector_parameter> parameters;
  /**
     Type (saved in parameters[0].get_int(0)[0]): 1=per_polygon
                                                  2=per_vertex
						  3=per_vertex_and_polygon
   */


  /// name of the mesh
  std::string name;

};



#endif 	    /* !MESH_OBJECT_H_ */
